
from factorio_blueprint_analyser import factorio, item, utils

# -----------------------------------------------------------
# Assembly machines recipe class
# -----------------------------------------------------------
DIFFICULTY = "normal"


def get_recipe(name):
    # Check that the recipe exists
    if name not in factorio.recipies:
        utils.warning(f"No recipe found for {name}")
        return None

    return Recipe(name)


class Recipe:
    def __init__(self, name) -> None:
        self.name = name

        factorio_recipe = factorio.recipies[name]

        # Get result item
        nb_item_output = factorio_recipe["result_count"] \
            if "result_count" in factorio_recipe else 1

        self.result = item.Item(name, nb_item_output)
        # TODO: deal with multiple results

        # Get ingredients
        self.ingredients = []

        ingredients = factorio_recipe[DIFFICULTY]["ingredients"] \
            if DIFFICULTY in factorio_recipe \
            else factorio_recipe["ingredients"]

        for ingredient in ingredients:
            # The recipe ingredients can have two formats:
            # - ['fast-transport-belt', 1]
            # - {'amount': 20, 'type': 'fluid', 'name': 'lubricant'}

            try:
                if isinstance(ingredient, list):
                    self.ingredients.append(
                        item.Item(ingredient[0],
                                  ingredient[1]))

                elif isinstance(ingredient, dict):
                    if "type" in ingredient and ingredient["type"] == "fluid":
                        # Fluid ingredient, we ignore it
                        continue

                    self.ingredients.append(
                        item.Item(ingredient["name"],
                                  ingredient["amount"],
                                  type=ingredient["type"]))
            except KeyError:
                utils.warning(f"Something went wrong with the recipe {name}")
                self.exists = False
                return

        # Get production time
        self.time = factorio_recipe["energy_required"] \
            if "energy_required" in factorio_recipe \
            else 1

    def ingredient_required(self, ingredient_name):
        for ingredient in self.ingredients:
            if ingredient.name == ingredient_name:
                return True

        return False

    def get_ingredient_nb(self, ingredient_name):
        for ingredient in self.ingredients:
            if ingredient.name == ingredient_name:
                return ingredient.amount

        return None

    def all_ingredients_required(self, given_ingredients):
        for ingredient in self.ingredients:
            if ingredient.name not in given_ingredients:
                return False

        return True

    def get_ingame_image_path(self):
        # Item image url
        # To display the entity game image, we will the images
        # from : https://wiki.factorio.com/Category:Game_images

        # We need to transform the entity name to a valid url
        entity_wiki_name = self.name.capitalize().replace('-', '_')

        return f"https://wiki.factorio.com/images/{entity_wiki_name}.png"

    def __str__(self) -> str:
        ingredients_str = ", ".join(
            [f"{item.name}: {item.amount}" for item in self.ingredients])
        return f"{self.name} [{ingredients_str}] => ({self.result})"



from factorio_blueprint_analyser import node as node_service, utils
from pyvis.network import Network as NetworkDisplay

# -----------------------------------------------------------
# Create a node network from a blueprint
# Provide calculations methods
# -----------------------------------------------------------


def create_network(blueprint):
    network_creator = NetworkCreator(blueprint)
    network = network_creator.create_network()

    # We give the network to the blueprint
    # it will be used to export the analysis
    blueprint.network = network
    return network


class NetworkCreator:
    def __init__(self, blueprint):
        self.blueprint = blueprint
        self.node_map = []

    def create_network(self):
        # Create a 2D array that will contain all nodes,
        # the same way as the blueprint array
        # Knowing where the nodes are located from each other will be useful

        # The nodes will then be exctracted from the 2D array in a list

        self.node_map = []

        for _ in range(self.blueprint.heigth):
            self.node_map += [[None for _ in range(self.blueprint.width)]]

        # Iterate over the blueprint entities
        # to create each nodes recursively
        for y in range(self.blueprint.heigth):
            for x in range(self.blueprint.width):
                self.create_node(x, y)

        return Network(self.blueprint, self.node_map)

    def create_node(self, x, y):
        # Returns a node object or None

        # Check if the cell hasn't been filled yet
        if x < 0 or x >= self.blueprint.width or y < 0 or y >= self.blueprint.heigth:
            return None

        # Check if a node already exists in the cell
        if self.node_map[y][x] is not None:
            return self.node_map[y][x]

        # Get the entity at the given position
        entity = self.blueprint.array[y][x]

        if entity is None:
            return None

        # We can now create the node
        node = node_service.create_node(entity)

        # Each game entity interacts with the other nodes in there own way
        # The requiered nodes will be created recursively
        if node.type == "transport-belt":
            # The node is inserted in the map to avoid infinit recursion
            self.node_map[y][x] = node

            # We want to set the entity in front of the belt as the node's child
            # We get the coordinates of the entity in front of the belt:
            tile_in_front_offset = entity.get_tile_in_front_offset()
            target_x = x + tile_in_front_offset[0]
            target_y = y + tile_in_front_offset[1]

            # We get the node in front of the belt or create a new one
            child_node = self.create_node(target_x, target_y)

            if child_node is not None and entity.can_connect_to(child_node.entity):
                node.childs.append(child_node)
                child_node.parents.append(node)

            return node

        elif node.type == "inserter":
            # The node is inserted in the map to avoid infinit recursion
            self.node_map[y][x] = node

            # Set the entity where items are droped as the node's child
            tile_drop_offset = entity.get_drop_tile_offset()
            target_drop_x = x + tile_drop_offset[0]
            target_drop_y = y + tile_drop_offset[1]

            drop_child_node = self.create_node(target_drop_x, target_drop_y)

            if drop_child_node is not None and entity.can_move_to(drop_child_node.entity):
                node.childs.append(drop_child_node)
                drop_child_node.parents.append(node)

            # Set the entity where items are picked up as the node's parent
            tile_pickup_offset = entity.get_pickup_tile_offset()
            target_pickup_x = x + tile_pickup_offset[0]
            target_pickup_y = y + tile_pickup_offset[1]

            pickup_node = self.create_node(target_pickup_x, target_pickup_y)

            if pickup_node is not None and entity.can_move_from(pickup_node.entity):
                node.parents.append(pickup_node)
                pickup_node.childs.append(node)

            return node

        elif node.type == "assembling-machine":
            # We will create 9 nodes for the 9 tiles of the assembling machine
            # All the cells are the same object, they share their parents and childs
            for offset in node.entity.offsets:
                target_x = node.entity.position[0] + offset[0]
                target_y = node.entity.position[1] + offset[1]

                if target_x < 0 or target_x >= self.blueprint.width\
                        or target_y < 0 or target_y >= self.blueprint.heigth:
                    continue

                self.node_map[target_y][target_x] = node
            return node

        elif node.type == "underground-belt":
            self.node_map[y][x] = node

            if entity.belt_type == "output":
                # Set the entity where items are droped as the node's child
                # Exactly the same as for the transport belt
                tile_in_front_offset = entity.get_tile_in_front_offset()
                target_x = x + tile_in_front_offset[0]
                target_y = y + tile_in_front_offset[1]

                # We get the node in front of the belt or create a new one
                child_node = self.create_node(target_x, target_y)

                if child_node is not None and entity.can_connect_to(child_node.entity):
                    node.childs.append(child_node)
                    child_node.parents.append(node)
            else:
                # We try to connect to the output belt
                possible_output_coords = entity.get_possible_output_coords()
                for possible_coord in possible_output_coords:
                    child_node = self.create_node(
                        possible_coord[0], possible_coord[1])

                    if child_node is not None and \
                            child_node.entity.name == node.entity.name and \
                            child_node.entity.belt_type == "output":

                        node.childs.append(child_node)
                        child_node.parents.append(node)
                        break

                    # No output belt found, print warning?

            return node

        elif node.type in ["container", "logistic-container"]:
            # Those entities does not interact with others
            self.node_map[y][x] = node
            return node

        elif node.type == "splitter":

            # We need to add the second splitter tile to the map
            if x == entity.position[0] and y == entity.position[1]:
                # If we are the original splitter, we need to add the second splitter
                self.node_map[y][x] = node

                second_belt_offset = entity.get_second_belt_offset()
                second_node_x = x + second_belt_offset[0]
                second_node_y = y + second_belt_offset[1]

                if second_node_x >= 0 and second_node_x < self.blueprint.width and\
                        second_node_y >= 0 and second_node_y < self.blueprint.heigth:
                    self.node_map[second_node_y][second_node_x] = node
            else:
                # We create the original splitter instead
                return self.create_node(entity.position[0], entity.position[1])

            # Set the entity where items are droped as the node's child
            drop_tile_offsets = entity.get_drop_tile_offsets()
            for offset in drop_tile_offsets:
                target_drop_x = x + offset[0]
                target_drop_y = y + offset[1]

                drop_child_node = self.create_node(
                    target_drop_x, target_drop_y)

                if drop_child_node is not None and entity.can_move_to(drop_child_node.entity):
                    node.childs.append(drop_child_node)
                    drop_child_node.parents.append(node)

            return node

        utils.warning(f"Unsupported entity type: {entity.data['type']}")
        return None


class Network:
    def __init__(self, blueprint, nodes_array):
        self.blueprint = blueprint
        self.nodes_array = nodes_array

        self.nodes = []

        for row in self.nodes_array:
            for node in row:
                if node is not None and not node.removed:
                    # Check that the node is not already in the list
                    # It's normal if the node takes multiple tiles
                    # (They appear multiple times in the map)
                    if node not in self.nodes:
                        self.nodes.append(node)

        self.optimize()

    def optimize(self):
        # We save the nodes original parents and childs
        for node in self.nodes:
            node.original_parents = [
                parent.entity.number for parent in node.parents]
            node.original_childs = [
                child.entity.number for child in node.childs]

        # Network optimisation
        for node in self.nodes:
            node.optimize()

        # Filter the removed nodes
        optimized_nodes = []

        for node in self.nodes:
            if not node.removed:
                optimized_nodes.append(node)

        self.nodes = optimized_nodes

    def get_node(self, entity_number) -> node_service.Node:
        for node in self.nodes:
            if node.entity.number == entity_number:
                return node

        return None

    def root_nodes(self):
        roots = []
        for node in self.nodes:
            if len(node.parents) == 0:
                roots.append(node)
        return roots

    def leaf_nodes(self):
        leafs = []
        for node in self.nodes:
            if len(node.childs) == 0:
                leafs.append(node)
        return leafs

    def calculate_bottleneck(self):
        # ==========================================
        # ====== Step 1: Purpose estimation ========
        # ==========================================

        # The first step is to calculate the purpose of each node
        # We will start from each assembling machine and go up and down
        # each parent and child node to tell them what we expect them to do

        # The purpose calculation starts from the assembly machines
        # childs as only one item is produced per assembling machine

        for node in self.nodes:
            if node.type == "assembling-machine":
                node.calculate_childs_purpose()

        # We will then, for the parents, deal with the assembling machines
        # with recipes that have one ingredient first as they are easier to process

        # Recipes with one ingredient first
        for node in self.nodes:
            if node.type == "assembling-machine" and\
                node.entity.recipe is not None and\
                    len(node.entity.recipe.ingredients) == 1:
                node.calculate_parents_purpose()

        # Recipes with multiple ingredients
        for node in self.nodes:
            if node.type == "assembling-machine" and\
                node.entity.recipe is not None and\
                    len(node.entity.recipe.ingredients) > 1:
                node.calculate_parents_purpose()

        # Display some debug info
        nb_transport_nodes_with_no_purpose = 0
        nb_transport_nodes = 0

        for node in self.nodes:
            if node.node_type == "transport_node":
                nb_transport_nodes += 1
                if node.transported_items is None or\
                        len(node.transported_items) == 0:
                    nb_transport_nodes_with_no_purpose += 1

        utils.verbose("")
        utils.success("Entities calculation estimation complete:")
        utils.verbose(
            f"{nb_transport_nodes - nb_transport_nodes_with_no_purpose} / {nb_transport_nodes} nodes with purpose")

        # ==============================================
        # ====== Step 2: Bottleneck calculation ========
        # ==============================================

        # We will try to estimate the use rate of all nodes
        # We start with the leaf nodes that have a purpose and
        # we will ask for the maximum produced item per second

        for node in self.leaf_nodes():
            items_output = node.get_materials_output()

            if items_output is None or len(items_output) == 0:
                continue

            # We ask each item that the node gives
            # If there is some capacity left with the item n,
            # we will ask for the item n+1 at the remaining capacity

            flow_capacity = node.entity.speed if node.entity.speed is not None else 10000

            for item_output in items_output:
                # We pass the flow to the node, it will send it to his childs
                acepted_amount = node.ask_flow(item_output.name, flow_capacity)
                flow_capacity -= acepted_amount

        utils.verbose("")
        utils.success("Bottleneck calculation complete!")
        utils.verbose("Produced items:")
        for node in self.leaf_nodes():
            # TODO: fix that nothing is shown for the bp blueprints4/drillFac1
            # None of the leaf nodes have a flow and none of them are processed by
            # the bottleneck algorithm (no node with 0 childs processed).
            # due to bp optimization ?
            for item in node.flow.items:
                utils.verbose(f"   {item}: {node.flow.items[item]} /s")

    def display(self):
        # Display the network as a node graph using the
        # networkx library
        net = NetworkDisplay(directed=True, height=1000, width=1900)
        net.repulsion(node_distance=80, spring_length=0)

        # Nodes and edges
        for node in self.nodes:
            # Define node size
            node_size = 3
            if node.type == "transport-belt":
                node_size = 3
            if node.type == "underground-belt" or "splitter":
                node_size = 4
            if node.type == "container" or "logistic-container":
                node_size = 3
            if node.type == "assembling-machine":
                node_size = 5
            if node.type == "inserter":
                node_size = 4

            # Display the node total items / second and the node use percentage
            node_label = ""
            bottleneck = False
            if node.node_type == "transport_node":
                node_label = str(
                    int(node.flow.total_amount * 100) / 100) + "/s "

                if node.usage_ratio is not None:
                    # Belts, arms, ...
                    node_label += str(int(node.usage_ratio * 100)) + "%"
                    bottleneck = int(node.usage_ratio * 100) >= 100

            net.add_node(node.entity.number,
                         value=node_size,
                         label=node_label,
                         shape="circularImage",
                         borderWidth=10,
                         color="lightgrey" if not bottleneck else "black",
                         image=node.entity.get_ingame_image_path(),
                         brokenImage="https://wiki.factorio.com/images/Warning-icon.png")

        for node in self.nodes:
            for child in node.childs:
                net.add_edge(node.entity.number,
                             child.entity.number,
                             value=1,
                             color="lightgrey")

        # Display recipes
        for node in self.nodes:
            if node.type == "assembling-machine" and node.entity.recipe is not None:
                node_id = str(node.entity.number) + "_recipe"

                # Display the produced items per sec
                items_per_sec = int(node.flow.total_amount * 100) / 100

                # Display the usage rate
                usage_rate = (str(int(node.usage_ratio * 100)) +
                              "%") if node.usage_ratio is not None else ""

                net.add_node(node_id,
                             label=str(items_per_sec) + "/s " + usage_rate,
                             value=3,
                             shape="image",
                             image=node.entity.recipe.get_ingame_image_path(),
                             brokenImage="https://wiki.factorio.com/images/Warning-icon.png")

                net.add_edge(node_id,
                             node.entity.number,
                             title="produce",
                             color="grey",
                             dashes=True,
                             arrowStrikethrough=False)

        # Display inputs
        for node in self.root_nodes():
            # TODO: Display the expected input materials
            node_id = str(node.entity.number) + "_root"
            net.add_node(node_id,
                         label="Input",
                         value=3,
                         shape="text")

            net.add_edge(node_id,
                         node.entity.number,
                         color="red",
                         dashes=True,
                         arrowStrikethrough=False)

        # Display outputs
        for node in self.leaf_nodes():
            # TODO: Display the expected produced materials
            node_id = str(node.entity.number) + "_leaf"
            net.add_node(node_id,
                         label="Output",
                         value=3,
                         shape="text")

            net.add_edge(node.entity.number,
                         node_id,
                         color="blue",
                         size=2,
                         dashes=True,
                         arrowStrikethrough=False)

        # Display nodes transported items and flow
        for node in self.nodes:
            if node.node_type == "transport_node" and node.transported_items is not None:
                for (i, item) in enumerate(node.transported_items):
                    node_id = str(node.entity.number) + "_item_" + str(i)

                    node_label = " "
                    if len(node.transported_items) > 1:
                        # We display the flow if there is one
                        if item.name in node.flow.items:
                            node_label = str(
                                int(node.flow.items[item.name] * 100) / 100) + "/s"

                    net.add_node(node_id,
                                 label=node_label,
                                 value=2,
                                 shape="image",
                                 image=item.get_ingame_image_path(),
                                 brokenImage="https://wiki.factorio.com/images/Warning-icon.png")

                    net.add_edge(node_id,
                                 node.entity.number,
                                 title="transport",
                                 color="lightgrey")

        # Display the graph
        net.show(self.blueprint.label.replace('/', '-') + ".html")




from factorio_blueprint_analyser import (
    factorio,
    blueprint,
    network,
    config
)

def init(config_dict=None):
    # Init config
    config.load_config(config_dict)

    # Load the Factorio data
    factorio.load_data()


def analyse_blueprint(blueprint_string):
    bp = blueprint.load_blueprint(blueprint_string)
    return _process_blueprint(bp)


def analyse_blueprint_from_path(blueprint_path):
    bp = blueprint.load_blueprint_from_path(blueprint_path)
    return _process_blueprint(bp)


def _process_blueprint(bp):
    bp.display()

    # Creade a node network from the blueprint
    nw = network.create_network(bp)

    # Calculate bottleneck
    nw.calculate_bottleneck()
    if config.config.display_network:
        nw.display()

    # Export the analysis
    analysis_result = bp.get_analysis()

    return analysis_result



import zlib
import json
import sys
import base64
from termcolor import colored

from factorio_blueprint_analyser import config


def verbose(content, end="\n", level=3):
    # Verbose level
    # 0: no output
    # 1: only errors
    # 2: errors and warnings
    # 3: errors, warnings and info
    if config.config.verbose_level >= level:
        print(content, end=end, file=sys.stderr, flush=True)


def warning(content):
    verbose(colored(f"WARNING: {content}", "yellow"), level=2)


def success(content):
    verbose(colored(f"SUCCESS: {content}", "green"), level=3)


def decode(string):
    return json.loads(zlib.decompress(
        base64.b64decode(string[1:])).decode('utf8'))


def encode(dict):
    return '0' + base64.b64encode(zlib.compress(bytes(json.dumps(dict), 'utf8'))).decode('utf8')



import os
import pathlib

config = None

# Finding where the assets are for when this is installed as a package
parent_path = pathlib.Path(__file__).parent.resolve()

# factorio:
#   # Inserter_capacity_bonus
#   # number between 0 and 7
#   # (https://wiki.factorio.com/Inserter_capacity_bonus_(research))
#   inserter_capacity_bonus: 0

#   # The facorio data generated from the game files
#   # This default data corresponds to the vanilla game
#   # To set your own data, you can follow our mod guide
#   # (Comming soon)
#   data_file_path: "factorio_blueprint_analyser/assets/factorio_raw/factorio_raw_min.json"

# network:
#   # The alogrithm will displat the
#   # results on a web page in a node network
#   # You can disable this feature by setting this to false
#   display: true

# # Verbose level
# # 1: only errors
# # 2: errors and warnings
# # 3: errors, warnings and info
# verbose_level: 3


class Config:
    # Factorio
    inserter_capacity_bonus = 0
    data_file_path = str(parent_path) + \
        "/assets/factorio_raw/factorio_raw_min.json"
    # Network
    display_network = True
    # Verbose level
    verbose_level = 3

    def __init__(self, config=None):
        if config:
            if "inserterCapacityBonus" in config:
                bonus = config["inserterCapacityBonus"]
                if type(bonus) is int and 0 <= bonus <= 7:
                    self.inserter_capacity_bonus = bonus
                else:
                    print(
                        f"Config warning: Invalid inserterCapacityBonus value: {bonus}. The value must be an \
                        integer between 0 and 7. Using default value: {self.inserter_capacity_bonus}")

            if "dataFilePath" in config:
                path = config["dataFilePath"]
                if os.path.exists(path):
                    self.data_file_path = path
                else:
                    print(
                        f"Config warning: Invalid dataFilePath value: {path}. The file does not exist.\
                             Using default value: {self.data_file_path}")

            if "displayNetwork" in config:
                if type(config["displayNetwork"]) is bool:
                    self.display_network = config["displayNetwork"]
                else:
                    print(
                        f"Config warning: Invalid displayNetwork value: {config['displayNetwork']}. \
                            The value must be a boolean. Using default value: {self.display_network}")

            if "verboseLevel" in config:
                level = config["verboseLevel"]
                if type(level) is int and 0 <= level <= 3:
                    self.verbose_level = level
                else:
                    print(
                        f"Config warning: Invalid verboseLevel value: {level}. The value must be an \
                        integer between 0 and 3. Using default value: {self.verbose_level}")


def load_config(config_dict):
    global config
    config = Config(config_dict)



import json

from factorio_blueprint_analyser import utils, entity, config

# -----------------------------------------------------------
# Read the blueprint from the given file
# Decode the file is encoded
# Create an entity list from the blueprint items
# Place the entities in a 2D array according to their position
# -----------------------------------------------------------


class Blueprint:
    entities = []
    label = ""
    heigth = 0
    width = 0
    array = []
    blueprint = None

    def __init__(self, bp_json):
        self.blueprint = bp_json

        # Check if the json is valid
        if "blueprint_book" in bp_json:
            if not "blueprints" in bp_json["blueprint_book"]:
                raise Exception("Invalid blueprint book, no blueprints found")

            if len(bp_json["blueprint_book"]["blueprints"]) == 0:
                raise Exception("Invalid blueprint book, no blueprints found")

            bp_json = bp_json["blueprint_book"]["blueprints"][0]
            self.blueprint = bp_json
            utils.warning(f"Blueprint book detected, using first blueprint")


        if "blueprint" not in bp_json:
            raise Exception("Invalid blueprint, no 'blueprint' key found")

        if "entities" not in bp_json["blueprint"]:
            # raise Exception("Invalid blueprint, no 'entities' key found")
            bp_json["blueprint"]["entities"] = []

        self.label = "No label"
        if "label" in bp_json["blueprint"]:
            self.label = bp_json["blueprint"]["label"]

        entities = bp_json["blueprint"]["entities"]

        # === Blueprint pre process ===

        # entities creation
        for entity_dic in entities:

            # Replace infinit chests with a normal chest
            if entity_dic["name"] == "infinity-chest":
                entity_dic["name"] = "iron-chest"

            # Creation of the entity from the blueprint entity dictionary
            new_entity = entity.create_entity(entity_dic)

            if new_entity is not None:
                self.entities.append(new_entity)

        if len(self.entities) == 0:
            utils.warning(f"No entities in the blueprint {self.label}")
            return

        # For some reason, the blueprint does not always start at 0,0 so we set a new origin:
        lowest_x = min(e.position[0] for e in self.entities)
        lowest_y = min(e.position[1] for e in self.entities)

        for entity_obj in self.entities:
            entity_obj.position[0] -= lowest_x
            entity_obj.position[1] -= lowest_y

        self.width = max(e.position[0] for e in self.entities) + 1
        self.heigth = max(e.position[1] for e in self.entities) + 1

        # === Convertion of the blueprint into a 2D array ===
        self.array = []

        for _ in range(self.heigth):
            self.array.append([None] * self.width)

        for created_entity in self.entities:
            if created_entity.large:
                # Because entites like the assembling machine is a 3x3 block,
                # we need to store it in the array 9 times
                for offset in created_entity.offsets:
                    offset_coord_x = offset[0] + created_entity.position[0]
                    offset_coord_y = offset[1] + created_entity.position[1]

                    # check that the offset is in the bondaries
                    if offset_coord_x >= self.width or offset_coord_x < 0:
                        continue

                    if offset_coord_y >= self.heigth or offset_coord_y < 0:
                        continue

                    self.array[offset_coord_y][offset_coord_x] = created_entity

            else:
                self.array[created_entity.position[1]
                           ][created_entity.position[0]] = created_entity

        # === Post process ===

        # Adding a temporary entity to the array where arms pickup or drop items
        # on an empty tile

        for y in range(self.heigth):
            for x in range(self.width):
                e = self.array[y][x]
                if e is not None and e.data["type"] == "inserter":
                    # Check drop tile
                    # If the drop tile is empty, add a virtual chest
                    # No need to check the pickup tile because the inserter
                    # will pickup the item from the virtual chest
                    # created by the other arm
                    drop_coord = e.get_drop_tile_coord()
                    if self.is_coord_in_boundaries(drop_coord):
                        drop_tile = self.array[drop_coord[1]][drop_coord[0]]
                        if drop_tile is None:
                            # The drop tile is empty, we add a temporary entity
                            container = entity.create_entity(
                                {
                                    'entity_number': str(e.number) + "_virtual\n(empty tile)",
                                    'name': 'wooden-chest',
                                    'position': {'x': drop_coord[0], 'y': drop_coord[1]}
                                }, virtual=True)

                            utils.verbose(
                                f"Adding temporary entity {container}")
                            self.array[drop_coord[1]
                                       ][drop_coord[0]] = container

        utils.success(f"Blueprint {self.label} loaded successfully")

    def is_coord_in_boundaries(self, coord):
        return coord[0] >= 0 and coord[0] < self.width and\
            coord[1] >= 0 and coord[1] < self.heigth

    def display(self):
        if config.config.verbose_level < 3:
            return

        utils.verbose(
            f"\n{self.label}, width: {self.width}, heigth: {self.heigth}, {len(self.entities)} entities:")
        for entity in self.entities:
            utils.verbose("  " + str(entity))

        utils.verbose("")
        for (y, row) in enumerate(self.array):
            utils.verbose("   ", end=" ")
            for (x, entity) in enumerate(row):
                if entity is None:
                    utils.verbose(" ", end=" ")
                else:
                    if entity.large:
                        utils.verbose(entity.to_char([x, y]), end=" ")
                    else:
                        utils.verbose(entity.to_char(), end=" ")
            utils.verbose("")
        utils.verbose("")

    def get_analysis(self):
        # Read the blueprint network if it exists
        # Then fill the information in a new dictionary

        if self.network is None:
            raise Exception("No network found")

        # Base blueprint format :
        # "blueprint": {
        #     "icons": [...],
        #     "entities": [
        #         {
        #             "entity_number": 1,
        #             "name": "transport-belt",
        #             "position": {...},
        #             "direction": 2
        #         },
        #         {
        #             "entity_number": 2,
        #             "name": "assembly-machine-1",
        #             "position": {...},
        #             "direction": 2
        #         },
        #         {
        #             "entity_number": 3,
        #             "name": "transport-belt",
        #             "position": {...}
        #         }
        #     ],
        #     "item": "blueprint",
        #     "label": "beltFac1",
        #     "version": 281479275544576
        # }

        # Analysed blueprint format:
        # "blueprint": {
        #     "icons": [...],
        #     "entities": [
        #         {
        #             "entity_number": 1,
        #             "name": "transport-belt",
        #             "position": {...},
        #             "direction": 2,
        #
        #             "input": True,
        #             "transpoted_items": [{"iron-plate": 1.26}],
        #             "usage_rate": 0.87,
        #             "parents": [2],
        #             "childrens": [3, 4]
        #         },
        #         {
        #             "entity_number": 2,
        #             "name": "assembly-machine-1",
        #             "position": {...},
        #             "recipe": "transport-belt"
        #
        #             "usage_rate": 0.84,
        #             "parents": [...],
        #             "childrens": [...]
        #         },
        #         {
        #             "entity_number": 3,
        #             "name": "transport-belt",
        #             "position": {...},
        #
        #             "output": True
        #             "transpoted_items": [{"transport-belt": 0.84}],
        #             "usage_rate": 0.52,
        #             "parents": [...],
        #             "childrens": [...]
        #         }
        #     ],
        #     "item": "blueprint",
        #     "label": "beltFac1",
        #     "version": 281479275544576,
        #
        #     "items_input": [{"iron-plate": 1.26}],
        #     "items_output": [{"transport-belt": 0.84}],
        #     "entities_input": [1, 2, 3, ...],
        #     "entities_output": [32, 33, 34, ...],
        #     "entities_bottleneck": [18, 23],
        # }

        analysed_bp = self.blueprint.copy()

        # Pre load network input and output
        root_entities_number = [
            node.entity.number for node in self.network.root_nodes()]
        leaf_entities_number = [
            node.entity.number for node in self.network.leaf_nodes()]

        entities_bottleneck = []

        # Entities related information
        for entity in analysed_bp["blueprint"]["entities"]:
            node = self.network.get_node(entity["entity_number"])

            if node is None:
                continue

            # Adding usage_rate
            usage_rate = node.usage_ratio
            if usage_rate is not None:
                entity["usage_rate"] = usage_rate

                # If the node as been "compacted" with other entities
                # due to optimization, we need to update the oser entites
                for compacted_node in node.compacted_nodes:
                    compacted_entity = self._get_entity(
                        compacted_node.entity.number, analysed_bp["blueprint"]["entities"])
                    compacted_entity["usage_rate"] = usage_rate

                    # Adding bottleneck entities number
                    if usage_rate >= 1:
                        entities_bottleneck.append(
                            compacted_entity["entity_number"])

                # Adding bottleneck entities number
                if usage_rate >= 1:
                    entities_bottleneck.append(entity["entity_number"])

            # Adding input/output
            if node.entity.number in root_entities_number:
                entity["input"] = True

                for compacted_node in node.compacted_nodes:
                    compacted_entity = self._get_entity(
                        compacted_node.entity.number, analysed_bp["blueprint"]["entities"])
                    compacted_entity["input"] = True

            if node.entity.number in leaf_entities_number:
                entity["output"] = True

                for compacted_node in node.compacted_nodes:
                    compacted_entity = self._get_entity(
                        compacted_node.entity.number, analysed_bp["blueprint"]["entities"])
                    compacted_entity["output"] = True

            # Adding transpoted_items
            entity["transpoted_items"] = node.flow.items
            for compacted_node in node.compacted_nodes:
                compacted_entity = self._get_entity(
                    compacted_node.entity.number, analysed_bp["blueprint"]["entities"])
                compacted_entity["transpoted_items"] = node.flow.items

            # Adding parents and childrens
            entity["parents"] = node.original_parents
            entity["children"] = node.original_childs

            for compacted_node in node.compacted_nodes:
                compacted_entity = self._get_entity(
                    compacted_node.entity.number, analysed_bp["blueprint"]["entities"])

                compacted_entity["parents"] = compacted_node.original_parents
                compacted_entity["children"] = compacted_node.original_childs

        # Blueprint related information
        # Adding the total in and out flow
        items_input = []
        for root_node in self.network.root_nodes():
            items_input.append(root_node.flow.items)

        items_output = []
        for leaf_node in self.network.leaf_nodes():
            items_output.append(leaf_node.flow.items)

        analysed_bp["items_input"] = items_input
        analysed_bp["items_output"] = items_output

        # Adding the entities input and output
        analysed_bp["entities_input"] = root_entities_number
        analysed_bp["entities_output"] = leaf_entities_number

        # Adding the entities bottleneck
        analysed_bp["entities_bottleneck"] = entities_bottleneck

        return analysed_bp

    def _get_entity(self, entity_number, entites):
        for entity in entites:
            if entity["entity_number"] == entity_number:
                return entity

        return {}


def load_blueprint(blueprint_sting):
    try:
        # Try to read the string directly as a JSON
        blueprint_json = json.loads(blueprint_sting)
    except json.decoder.JSONDecodeError:
        # If it fails, try to decode it
        blueprint_json = utils.decode(blueprint_sting)

    return Blueprint(blueprint_json)


def load_blueprint_from_path(file_path):
    # Read the file
    if file_path.endswith(".json"):
        # No need to decode the json
        with open(file_path, 'r') as f:
            bp_json = json.load(f)

    else:
        with open(file_path, 'r') as f:
            bp_encoded = f.read()
        bp_json = utils.decode(bp_encoded)

    return Blueprint(bp_json)



from factorio_blueprint_analyser import utils, item

# -----------------------------------------------------------
# Network nodes properties
# -----------------------------------------------------------


def create_node(entity):
    if entity.data["type"] == "assembling-machine":
        return Assembly_node(entity)
    else:
        return Transport_node(entity)


class Node:
    def __init__(self, entity):
        # Network construction data
        self.entity = entity
        self.childs = []
        self.parents = []
        self.original_parents = []
        self.original_childs = []
        self.type = entity.data["type"]

        # Network optimization data
        self.removed = False
        self.compacted_nodes = []  # Contain the nodes deleted by the optimizer

        # Bottleneck calculation
        self.flow = item.Flow()

    # Optimization
    def optimize(self):
        # Optimize the graph by removing the node if it's not needed.

        if self.type == "transport-belt":
            # If the transport belt as a single belt parent with same name
            # and no childs, we can remove it
            if len(self.childs) == 0 and \
                    len(self.parents) == 1 and \
                    self.parents[0].entity.name == self.entity.name:

                self.remove()

            # If the transport belt as a single child
            # and a single belt parent with same name, we can remove it
            elif len(self.childs) == 1 and \
                    len(self.parents) == 1 and \
                    self.parents[0].entity.name == self.entity.name:

                self.remove()

    def remove(self):
        # Remove the node from the network
        if len(self.childs) > 1 or len(self.parents) != 1:
            raise Exception("Can't remove this node " + str(self))

        self.removed = True

        # We need to replace our parent child with our child
        self.parents[0].childs.remove(self)
        if len(self.childs) > 0:
            # We need to replace our child parent with our parent
            self.childs[0].parents.remove(self)
            if self.childs[0] != self.parents[0]:
                # This condition avoid belts loop to be removed
                self.childs[0].parents.append(self.parents[0])
                self.parents[0].childs.append(self.childs[0])

        # We keep a trace of this node by adding it to the compacted list
        self.compacted_nodes.append(self)
        self.parents[0].compacted_nodes += self.compacted_nodes

    # Purpose estimation
    def get_materials_output(self):
        # Get the materials output of the node
        # If the node is an assembling machine, the output is the recipe result
        # Else, the output is the node inputs
        return None

    def set_purpose_from_parent(self, items):
        pass

    def set_purpose_from_child(self, items):
        pass

    def connected_to_input(self):
        # True if no assembly mach between the node and the root
        if self.node_type == "assembling-machine":
            return False

        if len(self.parents) == 0:
            return True

        for parent in self.parents:
            if parent.connected_to_input():
                return True

        return False

    # Bottleneck calculation
    def get_materials_input(self):
        # Get the materials input of the node
        # If the node is an assembling machine, the input is the recipe inputs
        # Else, the input is the node outputs
        return None

    # Other
    def __str__(self):
        compacted_info = ""
        if len(self.compacted_nodes) > 0:
            compacted_info = "[⧈ " + str(len(self.compacted_nodes)) + "]"

        return f"{self.entity} [{len(self.parents)} ► {len(self.childs)}] {compacted_info}"


class Assembly_node (Node):
    def __init__(self, entity):
        super().__init__(entity)
        self.node_type = "assembly_node"

        # Purpose calculation data
        self.inputs = []
        self.outputs = []

        if self.entity.recipe is not None:
            # Set the self inputs as the recipe ingredients
            for input_item in self.entity.recipe.ingredients:
                self.inputs.append(input_item)

            # Set the self outputs as the recipe result
            # We only consider that the recipes makes one item at the moment
            # TODO: Add support for multiple items
            self.outputs = [self.entity.recipe.result]

    def __str__(self):
        inputs = ""
        for item in self.inputs:
            inputs += str(item) + " "

        outputs = ""
        for item in self.outputs:
            outputs += str(item) + " "

        return f"Assembly node, {super().__str__()} [⭨ {inputs} ⭧ {outputs}]"

    # Purpose estimation
    def calculate_childs_purpose(self):
        # Set our childrens purpose

        if self.entity.recipe is None:
            return  # No purpose

        # We just set the purpose of our childs according to the outputs of the recipe
        for child in self.childs:
            child.set_purpose_from_parent([self.entity.recipe.result])

    def calculate_parents_purpose(self):
        # Set our parents purpose

        if self.entity.recipe is None:
            return  # No purpose

        # We will assign our parents a purpose according to the recipe inputs and the
        # purpose our parents already have

        # If we have no parents, we are an input, there is nothing more to do
        if len(self.parents) == 0:
            return

        if len(self.entity.recipe.ingredients) == 1:
            # We only need one ingredient to make the recipe
            # so our parents purpose is to provide the ingredient

            for parent in self.parents:
                # Creating a copy of the recipe input to avoid modifying the recipe
                recipe_ingredients = self.entity.recipe.ingredients[:]
                parent.set_purpose_from_child(recipe_ingredients)

        else:
            # We need more than one ingredient to make the recipe,
            # but we don't know witch parent will provide which ingredient

            # === 1. Finding our parents purpose ===

            utils.verbose(
                f"Finding providers for recipe {self.entity.recipe.name} with ingredinents:")

            # So we start by getting all our parents output items
            parent_outputs = []
            for parent in self.parents:
                parent_output = parent.get_materials_output()
                parent_outputs.append(parent_output)

            # === 2. Finding if our ingredients are all available ===
            # for each of our ingredients, we try to find a parent
            # that provides the ingredient
            provided_ingredients = [None] * \
                len(self.entity.recipe.ingredients)

            for (ing_index, ingredient) in enumerate(self.entity.recipe.ingredients):
                utils.verbose(f"\t{ingredient}")
                provided_ingredients[ing_index] = []
                for (parent_ind, parent_output) in enumerate(parent_outputs):
                    for parent_item in parent_output:
                        if ingredient.name == parent_item.name:
                            # We found a parent that provides the ingredient
                            # We can ignore it and the ingredient it provides
                            provided_ingredients[ing_index].append(
                                parent_item)
                            utils.verbose(
                                f"\t\t{self.parents[parent_ind]} provides {ingredient}")

            nb_treated_ingredients = 0
            needed_ingredients = []
            for (i, provided_ingredient) in enumerate(provided_ingredients):
                if len(provided_ingredient) > 0:
                    nb_treated_ingredients += 1
                else:
                    needed_ingredients.append(
                        self.entity.recipe.ingredients[i])

            utils.verbose(
                f"\t  {nb_treated_ingredients}/{len(self.entity.recipe.ingredients)} provided ingredients")

            if nb_treated_ingredients == len(self.entity.recipe.ingredients):
                # We consider that all our ingredients are provided
                return

            # === 3. Assigning available parents to the missing ingredients ===

            # The parents with no purpose will provide the other ingredients
            parent_without_purpose = []
            for parent in self.parents:
                if parent.transported_items is None:
                    parent_without_purpose.append(parent)
                    utils.verbose(
                        f"\t\t{parent} has no purpose, it will provide the other ingredients")

            if len(parent_without_purpose) > 0:
                # Assign the other ingredients to the parents
                for parent in parent_without_purpose:
                    parent.set_purpose_from_child(needed_ingredients)
                return

            # === 4. Assigning root connected parents to the missing ingredients ===

            # No available parents can provide the other ingredients
            # So we will find the parents that are directly connected to an input
            # without passing through another assembly node

            for parent in self.parents:
                if parent.connected_to_input():
                    utils.verbose(
                        f"\t\t{parent} is directly linked to an input, it will provide the other ingredients")

                    parent.set_purpose_from_child(needed_ingredients)
                    return

            # No parent can provide the other ingredients
            utils.verbose(
                f"Waring No parent can provide the other ingredients for {self}", level=1)

    def get_materials_output(self):
        # Get the materials output of the node
        # For a assembling machine node, the output is the recipe result

        return self.outputs

    # Bottleneck calculation
    @property
    def usage_ratio(self):
        if self.entity.recipe is None:
            return None

        return self.flow.total_amount / self.entity.items_per_second

    def get_materials_input(self):
        # Get the materials input of the node
        # For a assembling machine node, the input is the recipe inputs

        if self.entity.recipe is None:
            return None

        return self.inputs

    def ask_flow(self, item_name, amount):
        # We receive a flow request from a child

        if self.entity.recipe is None:
            # If no recipe, we give no flow
            return 0

        # We check that we produce the requested item
        if self.entity.recipe.result.name != item_name:
            return 0

        # We check that we have enougth assembly time
        available_amount = self.entity.items_per_second - self.flow.total_amount

        if available_amount <= 0:
            # We don't have enough assembly time to produce the requested amount
            return 0

        amount_that_need_to_be_produced = min(amount, available_amount)
        usage = amount_that_need_to_be_produced / self.entity.items_per_second

        if len(self.parents) == 0:
            # If no parents, we are an input,
            # so we supose that we have all the elements we need
            self.flow.add_item(item_name, amount_that_need_to_be_produced)
            return amount_that_need_to_be_produced

        # We check that we have the needed ingredients
        # so we ask for each of our items how much our parents can provide
        ingredients_input = {}
        for item in self.entity.recipe.ingredients:
            ingredients_input[item.name] = {"total": 0, "from": []}

            # We need to get the amount of the item we need the parents to provide
            # First we get the amount of the item that we would have needed in case of a 100% usage
            required_item_per_second = self.entity.required_items_per_second[item.name]

            # Then we reduce it to correspond to the amount of the item we expect to produce
            required_item_per_second *= usage
            required_item_per_second_target = required_item_per_second

            # We then ask our parents for the item:
            for parent in self.parents:
                amount_provided = parent.ask_flow(
                    item.name, required_item_per_second)
                ingredients_input[item.name]["from"].append({
                    "amount": amount_provided, "from": parent})
                ingredients_input[item.name]["total"] += amount_provided
                required_item_per_second -= amount_provided

                if required_item_per_second <= 0:
                    break

            if ingredients_input[item.name]["total"] == 0:
                # We don't have the needed ingredient, we can't proceed
                usage = 0
                break

            percent_recieved = ingredients_input[item.name]["total"] / \
                required_item_per_second_target

            # If the percent of the item we received is less than 100%,
            # we have to adjust our next items requests because
            # we don't need as much as expected

            usage *= percent_recieved

        produced_amount = usage * self.entity.items_per_second
        self.flow.add_item(self.entity.recipe.result.name, produced_amount)

        # Before sending how much item we can produce
        # we need to update our parents flow

        for ingredient_name in ingredients_input:
            required_item_per_second = self.entity.required_items_per_second[ingredient_name]
            real_required_amount = required_item_per_second * usage
            if ingredients_input[ingredient_name]["total"] > real_required_amount:
                # We asked too much, we need to give back
                exceeding_amount = ingredients_input[ingredient_name]["total"] - \
                    real_required_amount

                for amount_from in ingredients_input[ingredient_name]["from"]:
                    parent = amount_from["from"]
                    reduced_flow = parent.take_back_flow(
                        ingredient_name, exceeding_amount)
                    exceeding_amount -= reduced_flow

                    if exceeding_amount <= 0:
                        break

        return produced_amount

    def take_back_flow(self, item_name, amount):
        # We gived too much flow at some point, we take it back

        if self.entity.recipe is None:
            return 0

        # We check that we produce the requested item
        if self.entity.recipe.result.name != item_name:
            return 0

        if self.flow.total_amount <= 0:
            return 0

        # We get only take back the flow we can take back
        amount_to_take_back = amount
        if amount > self.flow.total_amount:
            amount_to_take_back = self.flow.total_amount

        percentage_to_take_back = amount_to_take_back / self.flow.total_amount

        # We ask our parents to take back their flow
        for item in self.entity.recipe.ingredients:
            actual_required_amount = self.usage_ratio * \
                self.entity.required_items_per_second[item.name]

            item_amount_to_take_back = actual_required_amount * percentage_to_take_back

            for parent in self.parents:
                amount_taked_back = parent.take_back_flow(
                    item.name, item_amount_to_take_back)
                item_amount_to_take_back -= amount_taked_back

                if item_amount_to_take_back <= 0:
                    break

        self.flow.reduce(item_name, amount_to_take_back)
        return amount_to_take_back


class Transport_node (Node):
    def __init__(self, entity):
        super().__init__(entity)
        self.node_type = "transport_node"

        # Bottleneck calculation data
        self.transported_items = None

    def __str__(self):
        transported_items = ""

        if self.transported_items is None:
            transported_items = "?"

        elif len(self.transported_items) == 0:
            transported_items = "None"
        else:
            for item in self.transported_items:
                transported_items += str(item) + " "

        return f"Node {super().__str__()} [↔ {transported_items}]"

    # Purpose estimation
    def get_materials_output(self):
        # Get the materials output of the node

        if self.transported_items is None:
            # We don't know what the node outputs are
            # so we ask our parents for their output
            transported_items = []
            for parent in self.parents:
                transported_items += parent.get_materials_output()

            if len(transported_items) > 0:
                self.transported_items = transported_items

            return transported_items

        return self.transported_items

    def set_purpose_from_child(self, items):
        # Our childrens are telling us the items that they need
        self.add_transported_item(items)

        # We send the message to our parents
        for parent in self.parents:
            parent.set_purpose_from_child(items)

    def set_purpose_from_parent(self, items):
        # Our parents are telling us the items that they will give us
        self.add_transported_item(items)

        # We send the message to our childs
        for child in self.childs:
            child.set_purpose_from_parent(items)

    def add_transported_item(self, items):
        if self.transported_items is None:
            self.transported_items = items
        else:
            for item in items:
                item_already_in_list = False
                for transported_item in self.transported_items:
                    if transported_item.name == item.name:
                        # We already transport the item so we can ignore it
                        item_already_in_list = True
                        break

                if not item_already_in_list:
                    self.transported_items.append(item)

    def is_item_transported(self, item_name):
        if self.transported_items is None:
            return False

        for transported_item in self.transported_items:
            if transported_item.name == item_name:
                return True

        return False

    # Bottleneck calculation
    def get_materials_input(self):
        # Get the materials input of the node
        # For a transport node, the input is the transported items
        return self.transported_items

    @property
    def usage_ratio(self):
        if self.entity.speed is None:
            return None

        return self.flow.total_amount / self.entity.speed

    def ask_flow(self, item_name, amount):
        # An item flow is requiered from a parent node
        # We check that we tranport the requested item
        if not self.is_item_transported(item_name):
            return 0

        if self.entity.speed is None:
            # Node without speed or usage_ratio limits (chests)
            provided_amount = self.get_parents_flow(item_name, amount)
            self.flow.add_item(item_name, provided_amount)
            return provided_amount

        if self.usage_ratio >= 1:
            # We are full, we can't give more flow
            return 0

        # We get only the flow we can accept
        available_flow_amount = self.entity.speed - self.flow.total_amount

        processed_amount = amount
        if available_flow_amount < processed_amount:
            # We have an exceding flow,
            # we can't accept all of the second flow
            processed_amount = available_flow_amount

        # How much of this new flow can our children accept
        provided_amount = self.get_parents_flow(item_name, processed_amount)
        self.flow.add_item(item_name, provided_amount)
        return provided_amount

    def get_parents_flow(self, item_name, amount):
        # If we don't have parents, we are an input node
        if len(self.parents) == 0:
            # so we provide all the requested flow
            return amount

        # If we have one parent, we request from him the flow
        elif len(self.parents) == 1:
            return self.parents[0].ask_flow(item_name, amount)

        # If we have multiple parents,
        # We ask them each the flow, they are sorted by priority
        # TODO: sort by priority (arms and arms with higer speed first, ...)

        requested_amount = amount
        sendedable_amount = 0
        for parent in self.parents:
            provided_amount = parent.ask_flow(item_name, requested_amount)
            sendedable_amount += provided_amount
            # If there is some usage_ratio left afetr the previous parent,
            # we ask the next parent
            requested_amount -= provided_amount
            if requested_amount <= 0:
                break

        return sendedable_amount

    def take_back_flow(self, item_name, amount):
        # We gived too much flow at some point, we take it back
        if amount <= 0:
            return 0

        # We check that we tranport the requested item
        if not self.is_item_transported(item_name):
            return 0

        if self.entity.speed is None:
            # Node without speed or usage_ratio limits (chests)
            taked_back_amount = self.take_back_parents_flow(item_name, amount)
            self.flow.reduce(item_name, taked_back_amount)
            return taked_back_amount

        if self.usage_ratio <= 0:
            # We don't have flow to take back
            return 0

        # We get only take back the flow we can take back
        amount_to_take_back = amount
        if amount > self.flow.total_amount:
            amount_to_take_back = self.flow.total_amount

        # How much of this new flow can our children accept
        taked_back_amount = self.take_back_parents_flow(
            item_name, amount_to_take_back)
        self.flow.reduce(item_name, taked_back_amount)
        return taked_back_amount

    def take_back_parents_flow(self, item_name, amount):
        # Same as get_parents_flow but to take it back

        if len(self.parents) == 0:
            # Input node
            return amount

        elif len(self.parents) == 1:
            return self.parents[0].take_back_flow(item_name, amount)

        amount_to_take_back = amount
        taked_back_amount_total = 0
        for parent in self.parents:

            taked_back_amount = parent.take_back_flow(
                item_name, amount_to_take_back)
            taked_back_amount_total += taked_back_amount
            amount_to_take_back -= taked_back_amount
            if amount_to_take_back <= 0:
                break

        return taked_back_amount_total



import io
import argparse
import sys
import os

# -----------------------------------------------------------
# Read the user input and check the options
# The options are stored in the global variables
# -----------------------------------------------------------

input = ""
output = ""
force = False
config_path = ""


def read_options():
    global input, output, force, config_path

    # ==== Options read ====

    sys.stdin = io.TextIOWrapper(sys.stdin.detach(), encoding="utf-8")
    sys.stdout = io.TextIOWrapper(sys.stdout.detach(), encoding="utf-8")
    sys.stderr = io.TextIOWrapper(sys.stderr.detach(), encoding="utf-8")

    parser = argparse.ArgumentParser(
        description="Find the bottleneck in a Factorio blueprint")

    parser.add_argument("-i", "--input", nargs="?", dest="input",
                        help="Blueprint JSON or encoded file path", default="./examples/beltFac.json")

    parser.add_argument("-o", "--output", nargs="?", dest="output",
                        help="JSON File output for the analysed blueprint", default="analysed_blueprint.json")

    parser.add_argument("-f", "--force", action="store_true", dest="force",
                        help="Force overwrite of existing result file", default=False)

    parser.add_argument("-c", "--config", nargs="?", dest="config",
                        help="Analyser yaml config file path", default="config/config_default.yaml")

    opt = parser.parse_args()

    input = opt.input
    output = opt.output
    force = opt.force
    config_path = opt.config

    # ==== Options validation ====

    # Check if the input file exists
    if not os.path.exists(opt.input):
        raise Exception(f"Input file '{opt.input}' does not exist")

    # Check if the output file exists
    if os.path.exists(opt.output) and not force:
        raise Exception(
            f"Output file '{opt.output}' already exists\nUse --force or -f to overwrite it")

    # Check if the config file exists
    if not os.path.exists(opt.config):
        raise Exception(f"Config file '{opt.config}' does not exist")




# -----------------------------------------------------------
# Items class
# Used to calculate bottleneck
# -----------------------------------------------------------


class Item:
    def __init__(self, name, amount, type="item"):
        self.name = name
        self.amount = amount
        self.type = type

    def __str__(self):
        return f"{self.name} ({self.amount})"

    def get_ingame_image_path(self):
        # Item image url
        # To display the entity game image, we will the images
        # from : https://wiki.factorio.com/Category:Game_images

        # We need to transform the entity name to a valid url
        entity_wiki_name = self.name.capitalize().replace('-', '_')

        return f"https://wiki.factorio.com/images/{entity_wiki_name}.png"


class Flow:
    def __init__(self):
        self.items = {}
        # Format:  {
        #    item_name_1: 0.8, # Per min
        #    item_name_2: 0.4,
        # }

    def add_item(self, item, amount):
        if item is None:
            raise Exception("Flow added without item")

        if type(item) is not str:
            raise Exception("Flow added with invalid item")

        if amount is None:
            raise Exception("Flow added without amount")

        if item not in self.items:
            self.items[item] = amount
        else:
            self.items[item] += amount

    def reduce(self, item, amount):
        if item not in self.items:
            raise Exception("Item not in flow")

        self.items[item] -= amount
        if self.items[item] <= 0:
            del self.items[item]

    @property
    def total_amount(self):
        if len(self.items) == 0:
            return 0
        return sum(self.items.values())

    def __str__(self) -> str:
        return "[" + ", ".join([f"{item}: {amount}" for item, amount in self.items.items()]) + "]"



from math import floor
from termcolor import colored

from factorio_blueprint_analyser import factorio, recipe, utils, config

# -----------------------------------------------------------
# Base class for all entities
# The create_entity function will return an entity instance
# corresponding to the givenentity_in_blueprint
# -----------------------------------------------------------


def create_entity(entity_in_blueprint, virtual=False):
    # entity_in_blueprint examples:
    # {
    #     'entity_number': 18,
    #     'name': 'inserter',
    #     'position': {'x': 10, 'y': -90},
    #     'direction': 6
    # }
    # {
    #     "entity_number": 15,
    #     "name": "assembling-machine-1",
    #     "position": {
    #         "x": 3.5,
    #         "y": -90.5
    #     },
    #     "recipe": "iron-gear-wheel"
    # },

    # Check that the entity exists in the Factorio data
    if entity_in_blueprint["name"] not in factorio.entities:
        utils.warning(f"Entity {entity_in_blueprint['name']} not found in Factorio data")
        # sys.exit(1)
        return None

    entity_data = factorio.entities[entity_in_blueprint["name"]]

    # Virtual entities are entities that are not in the original blueprint
    # We have created them to solve certain edge cases such as the inserter
    # that needs to pickup the item from an empty tile
    entity_data["virtual"] = virtual

    # Return the corresponding Entity object
    if entity_data["type"] == "transport-belt":
        return TransportBelt(entity_in_blueprint, entity_data)

    elif entity_data["type"] == "assembling-machine":
        return AssemblingMachine(entity_in_blueprint, entity_data)

    elif entity_data["type"] == "inserter":
        if entity_data["name"] == "long-handed-inserter":
            return RedArm(entity_in_blueprint, entity_data)
        elif entity_data["name"] == "stack-inserter":
            return StackInserter(entity_in_blueprint, entity_data)
        else:
            return Inserter(entity_in_blueprint, entity_data)

    elif entity_data["type"] in ["container", "logistic-container"]:
        return Container(entity_in_blueprint, entity_data)

    elif entity_data["type"] == "underground-belt":
        return UndergroundBelt(entity_in_blueprint, entity_data)

    elif entity_data["type"] == "splitter":
        return Splitter(entity_in_blueprint, entity_data)

    utils.warning(f"entity {entity_in_blueprint['name']} of type {entity_data['type']} not supported")


# Enities interfaces
class Entity:
    def __init__(self, entity_in_blueprint, entity_data):
        self.data = entity_data
        self.number = entity_in_blueprint["entity_number"]
        self.name = entity_in_blueprint["name"]
        self.large = False

        # Correc position
        self.original_position = entity_in_blueprint["position"]
        self.position = [
            floor(entity_in_blueprint["position"]["x"]),
            floor(entity_in_blueprint["position"]["y"])
        ]

        # Get the direction
        if "direction" in entity_in_blueprint:
            self.direction = entity_in_blueprint["direction"]
        else:
            self.direction = None

        self.speed = None

    def __str__(self):
        speed = ""
        if self.speed is not None:
            speed = f"{self.speed}/s"

        return f"{self.number} {self.name} [{self.position[0]}, {self.position[1]}] {self.to_char()} {speed}"

    def to_char(self):
        return '?'

    def get_ingame_image_path(self):
        # Item image url
        # To display the entity game image, we will the images
        # from : https://wiki.factorio.com/Category:Game_images

        # We need to transform the entity name to a valid url
        entity_wiki_name = self.data['name'].capitalize().replace('-', '_')

        return f"https://wiki.factorio.com/images/{entity_wiki_name}.png"


class LargeEntity(Entity):
    # Assembling machines, splitters, furnace, etc.
    def __init__(self, entity_in_blueprint, entity_data):
        super().__init__(entity_in_blueprint, entity_data)
        self.large = True
        self.offsets = []

    def to_char(self, coords=[0, 0]):
        return '?'

    def can_connect_to(self, entity):
        # Check if the entity can be connected to this entity
        # For example, a belt can be connected to an other belt
        return False


# Factorio entities:
class TransportBelt(Entity):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)

        # Saving speed of the belt
        if "speed" not in entity_data:
            utils.warning(f"{self.name} has no speed")
            self.tile_per_sec = 0.03125  # the tile_per_sec of the lvl1 transport belt
        else:
            self.tile_per_sec = entity_data["speed"]

        # Calculation of the belt item per second
        # 60 ticks / second
        # A tile is 4 items
        # There is two line on the belt
        # this will result, for the first belt, with an output of 15 item per second
        # This can be veryfied here: https://wiki.factorio.com/Belt_transport_system

        self.speed = self.tile_per_sec * 60 * 4 * 2

    def to_char(self):
        color = "white"
        if self.name == "transport-belt":
            color = "yellow"
        if self.name == "fast-transport-belt":
            color = "red"
        if self.name == "express-transport-belt":
            color = "blue"

        # ← ↑ → ↓
        if self.direction == 2:
            return colored("→", color)
        elif self.direction == 4:
            return colored("↓", color)
        elif self.direction == 6:
            return colored("←", color)
        elif self.direction == 0 or self.direction is None:
            return colored("↑", color)

        return colored("?", color)

    def get_tile_in_front_offset(self):
        # Returns an offset of the tile
        # in front of the conveyor belt
        #   example: [1, 0] if the conveyor is facing the right

        if self.direction == 2:
            return [1, 0]
        elif self.direction == 4:
            return [0, 1]
        elif self.direction == 6:
            return [-1, 0]
        elif self.direction == 0 or self.direction is None:
            return [0, -1]
        
        utils.warning(f"Unknown direction {self.direction} for {self.name}")
        return [0, 0]

    def can_connect_to(self, entity):
        if entity.data["type"] == "transport-belt":
            # A belt can be connected to another belt
            # except if they are facing each other
            if self.direction == 2 and entity.direction == 6 or \
                    self.direction == 6 and entity.direction == 2:
                return False
            elif self.direction == 4 and entity.direction is None or \
                    self.direction is None and entity.direction == 4:
                return False

            return True

        elif entity.data["type"] == "underground-belt" and \
                entity.belt_type == "input" or \
                entity.data["type"] == "splitter":
            # A belt can be connected to an underground belt
            # except if they are facing each other

            # return self.direction == entity.direction

            if self.direction == 2 and entity.direction == 6 or \
                    self.direction == 6 and entity.direction == 2:
                return False
            elif self.direction == 4 and entity.direction is None or \
                    self.direction is None and entity.direction == 4:
                return False

            # We are ignoring the complex transfers for now
            return True

        return False


class Inserter (Entity):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)

        # Saving speed of the inserter
        if "rotation_speed" not in entity_data:
            utils.warning(f"{self.name} has no rotation speed")
            self.rotation_speed = 0.014  # the rotation_speed of the lvl1 inserter
        else:
            self.rotation_speed = entity_data["rotation_speed"]

        # The rotation speed is the turn per tick
        # There is 60 ticks per second
        self.speed = self.rotation_speed * 60  # turn or items per second

        # Inserter capacity bonnus https://wiki.factorio.com/Inserter_capacity_bonus_(research)
        if config.config.inserter_capacity_bonus >= 7:
            self.speed *= 3
        elif config.config.inserter_capacity_bonus >= 2:
            self.speed *= 2

    def get_drop_tile_offset(self):
        # Returns an offset of the tile where items are dropped
        #   example: [1, 0] if the arm is facing the right
        # For some reason, the inserter is facing
        # the oposite direction compared to the belts
        if self.direction == 2:
            return [-1, 0]
        elif self.direction == 4:
            return [0, -1]
        elif self.direction == 6:
            return [1, 0]
        else:
            return [0, 1]

    def get_drop_tile_coord(self):
        # Returns the coordinaties of the tile where items are dropped
        pickup_tile_offset = self.get_drop_tile_offset()
        return [
            pickup_tile_offset[0] + self.position[0],
            pickup_tile_offset[1] + self.position[1]
        ]

    def get_pickup_tile_offset(self):
        # Returns the offset of the tile where items are picked up
        return [
            -self.get_drop_tile_offset()[0],
            -self.get_drop_tile_offset()[1]
        ]

    def get_pickup_tile_coord(self):
        # Returns the coordinates of the tile where items are picked up
        pickup_tile_offset = self.get_pickup_tile_offset()
        return [
            pickup_tile_offset[0] + self.position[0],
            pickup_tile_offset[1] + self.position[1]
        ]

    def can_move_to(self, entity):
        # The arm can move items to belts,
        # underground belts, chests or assembling machines
        available_types = ["transport-belt",
                           "underground-belt",
                           "container",
                           "logistic-container",
                           "assembling-machine",
                           "splitter"]

        if entity.data["type"] in available_types:
            return True

        return False

    def can_move_from(self, entity):
        # The arm can move items from belts,
        # underground belts, chests or assembling machines

        # It seams that the allowed entities
        # are the same as the can_move_to
        return self.can_move_to(entity)

    def to_char(self):
        color = "white"
        if self.name == "inserter":
            color = "yellow"
        if self.name == "long-handed-inserter":
            color = "red"
        if self.name == "fast-inserter":
            color = "blue"

        # ► ▼ ▲ ◄
        if self.direction == 2:
            return colored("◄", color)
        elif self.direction == 4:
            return colored("▲", color)
        elif self.direction == 6:
            return colored("►", color)
        else:
            return colored("▼", color)


class StackInserter  (Inserter):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)

        # Rewriting the speed
        self.speed = self.rotation_speed * 60  # turn or items per second

        # Inserter capacity bonnus https://wiki.factorio.com/Inserter_capacity_bonus_(research)
        capacity_multiplier = 2 + config.config.inserter_capacity_bonus

        if config.config.inserter_capacity_bonus >= 5:
            capacity_multiplier += 1
        if config.config.inserter_capacity_bonus >= 6:
            capacity_multiplier += 1
        if config.config.inserter_capacity_bonus >= 7:
            capacity_multiplier += 1

        self.speed *= capacity_multiplier


class RedArm (Inserter):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)

    def get_drop_tile_offset(self):
        return [e * 2 for e in super().get_drop_tile_offset()]

    def get_ingame_image_path(self):
        # This entity image path is an exception
        return "https://wiki.factorio.com/images/Long-handed_inserter.png"


class AssemblingMachine (LargeEntity):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)

        self.offsets = [
            [0, 0],
            [0, 1],
            [0, -1],
            [1, 1],
            [1, 0],
            [1, -1],
            [-1, 1],
            [-1, 0],
            [-1, -1],
        ]

        self.recipe = None
        if "recipe" in dictionary_entity:
            self.recipe = recipe.get_recipe(dictionary_entity["recipe"])

        if self.recipe is not None:
            # Saving speed of the assembling machine
            if "crafting_speed" not in entity_data:
                utils.warning(f"{self.name} has no crafting speed", level=1)
                self.speed = 0.5  # the speed of the assembling-machine-1
            else:
                self.speed = entity_data["crafting_speed"]

            time_per_item = self.recipe.time / self.speed
            self.items_per_second = self.recipe.result.amount / time_per_item

            # Define the required items per second
            # to have the assembling machine working at 100%
            self.required_items_per_second = {}
            for item in self.recipe.ingredients:
                self.required_items_per_second[item.name] = item.amount / \
                    time_per_item

    def get_usage_ratio(self, ingredients_amount):
        # Calculate the number of items produced per second
        # according to an ingredients amount dictionary

        if self.recipe is None or not self.recipe.all_ingredients_required(ingredients_amount.keys()):
            return 0

        lowest_ingredient_requierment_ratio = None
        for ingredient in ingredients_amount:

            completion_ratio = ingredients_amount[ingredient] / \
                self.required_items_per_second[ingredient]

            if lowest_ingredient_requierment_ratio is None or \
                    completion_ratio < lowest_ingredient_requierment_ratio:
                lowest_ingredient_requierment_ratio = completion_ratio

        return min(1.0, lowest_ingredient_requierment_ratio)

    def to_char(self, coords=None):
        color = "white"
        if self.name == "assembling-machine-2":
            color = "blue"
        if self.name == "assembling-machine-3":
            color = "yellow"

        if coords is None:
            return colored(self.recipe.name[0] if self.recipe is not None else "?", color)

        offset = [coords[0] - self.position[0],
                  coords[1] - self.position[1]]

        # The entity_offset is used to tell witch part
        # of the assembling machine we want to display
        # For example, if the entity is an assembling machine (3x3 size),
        # the offset will be [0, 0] for the center, [1, 1] for the top-right,
        # [1, -1] for the bottom-right and so on.

        # ┌─┐
        # │A│
        # └─┘

        if offset == [0, 0] or coords is None:
            return colored(self.recipe.name[0] if self.recipe is not None else "?", color)
        elif offset == [1, 1]:
            return colored("┘", color)
        elif offset == [1, -1]:
            return colored("┐", color)
        elif offset == [-1, 1]:
            return colored("└", color)
        elif offset == [-1, -1]:
            return colored("┌", color)
        elif offset == [0, 1] or offset == [0, -1]:
            return colored("─", color)
        else:
            return colored("│", color)

    def __str__(self):
        recipe_str = ""
        if self.recipe is not None:
            recipe_str = self.recipe.name

        return super().__str__() + " [" + recipe_str + "]"


class Container (Entity):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)

    def to_char(self):
        if self.name == "logistic-chest-passive-provider":
            return colored("⧈", "red")
        elif self.name == "logistic-chest-active-provider":
            return colored("⧈", "magenta")
        elif self.name == "logistic-chest-buffer":
            return colored("⧈", "green")
        elif self.name == "logistic-chest-requester":
            return colored("⧈", "cyan")
        elif self.name == "logistic-chest-storage":
            return colored("⧈", "yellow")

        return "⧈"

    def get_ingame_image_path(self):

        if self.name == "logistic-chest-passive-provider":
            return "https://wiki.factorio.com/images/Passive_provider_chest.png"
        elif self.name == "logistic-chest-active-provider":
            return "https://wiki.factorio.com/images/Active_provider_chest.png"
        elif self.name == "logistic-chest-buffer":
            return "https://wiki.factorio.com/images/Buffer_chest.png"
        elif self.name == "logistic-chest-requester":
            return "https://wiki.factorio.com/images/Requester_chest.png"
        elif self.name == "logistic-chest-storage":
            return "https://wiki.factorio.com/images/Storage_chest.png"

        return super().get_ingame_image_path()


class UndergroundBelt (TransportBelt):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)

        self.belt_type = dictionary_entity["type"]  # "input" or "output"

        # Saving belt distance
        if "max_distance" not in entity_data:
            self.max_distance = 5  # the distance of the lvl1 underground-belt
        else:
            self.max_distance = entity_data["max_distance"]

        # Saving speed
        if "speed" not in entity_data:
            utils.warning(f"{self.name} has no speed")
            self.speed = 0.03125  # the speed of the lvl1 underground-belt
        else:
            self.speed = entity_data["speed"]

        self.speed = self.speed * 60 * 4 * 2

    def get_possible_output_coords(self):
        start_coord = self.position
        possible_coords = []

        for _ in range(self.max_distance):
            start_coord = [
                start_coord[0] + self.get_tile_in_front_offset()[0],
                start_coord[1] + self.get_tile_in_front_offset()[1]
            ]

            possible_coords.append(start_coord)

        return possible_coords

    def to_char(self):
        color = "white"
        if self.name == "underground-belt":
            color = "yellow"
        if self.name == "fast-underground-belt":
            color = "red"
        if self.name == "express-underground-belt":
            color = "blue"

        if self.belt_type == "input":
            # ⇐ ⇑ ⇒ ⇓
            if self.direction == 2:
                return colored("⇒", color)
            elif self.direction == 4:
                return colored("⇓", color)
            elif self.direction == 6:
                return colored("⇐", color)
            else:
                return colored("⇑", color)
        else:
            # ⇦ ⇨ ⇧ ⇩
            if self.direction == 2:
                return colored("⇨", color)
            elif self.direction == 4:
                return colored("⇩", color)
            elif self.direction == 6:
                return colored("⇦", color)
            else:
                return colored("⇧", color)

    def __str__(self):
        return super().__str__() + " " + self.belt_type


class Splitter (LargeEntity):
    def __init__(self, dictionary_entity, entity_data):
        super().__init__(dictionary_entity, entity_data)
        self.offsets = [[0, 0], self.get_second_belt_offset()]
        # TODO: Filters

        # Saving speed
        if "speed" not in entity_data:
            utils.warning(f"{self.name} has no speed")
            self.speed = 0.03125  # the speed of the lvl1 splitter
        else:
            self.speed = entity_data["speed"]

        self.speed = self.speed * 60 * 4 * 2

    def get_second_belt_offset(self):

        if self.direction == 2:  # ⇨
            return [0, -1]
        elif self.direction == 4:  # ⇩
            return [-1, 0]
        elif self.direction == 6:  # ⇦
            return [0, -1]
        else:  # ⇧
            return [-1, 0]

    def get_drop_tile_offsets(self):
        if self.direction == 2:  # ⇨
            return [[1, -1], [1, 0]]
        elif self.direction == 4:  # ⇩
            return [[-1, 1], [0, 1]]
        elif self.direction == 6:  # ⇦
            return [[-1, -1], [-1, 0]]
        else:  # ⇧
            return [[-1, -1], [0, -1]]

    def can_move_to(self, entity):
        if entity.data["type"] == "transport-belt":
            # A belt can be connected to another belt
            # except if they are facing each other
            if self.direction == 2 and entity.direction == 6 or \
                    self.direction == 6 and entity.direction == 2:
                return False
            elif self.direction == 4 and entity.direction is None or \
                    self.direction is None and entity.direction == 4:
                return False

            return True

        elif entity.data["type"] == "underground-belt" and \
                entity.belt_type == "input" or \
                entity.data["type"] == "splitter":
            # A belt can be connected to an underground belt
            # if they are in the same direction

            return self.direction == entity.direction

        return False

    def to_char(self, coords=None):

        color = "white"
        if self.name == "splitter":
            color = "yellow"
        if self.name == "fast-splitter":
            color = "red"
        if self.name == "express-splitter":
            color = "blue"

        # ⬑⬏ ⬐⬎
        # ↱    ↰
        # ↳    ↲

        if coords is None:
            return colored("⬑⬏", color)

        if self.position[0] == coords[0] and self.position[1] == coords[1]:
            if self.direction == 2:
                return colored("↳", color)
            elif self.direction == 4:
                return colored("⬎", color)
            elif self.direction == 6:
                return colored("↲", color)
            else:
                return colored("⬏", color)
        else:
            if self.direction == 2:
                return colored("↱", color)
            elif self.direction == 4:
                return colored("⬐", color)
            elif self.direction == 6:
                return colored("↰", color)
            else:
                return colored("⬑", color)


# ↕ ↔
# ╔═╗
# ║ ║
# ╚═╝

# ⭦ ⭧ ⭩ ⭨



from setuptools import setup

setup(
    name='factorioBlueprintAnalyser',
    version='1.3.0',
    description="A python library analyse Factorio Blueprints and find bottlenecks.",
    url="https://github.com/tomansion/factorio_blueprint_analyser_app/",
    author="Tom Mansion",
    license="MIT License",
    install_requires=["termcolor==2.0.1", "PyYAML==5.4.1", "pyvis==0.3.0"],
    package_data={
        "factorio_blueprint_analyser": [
            "assets/factorio_raw/factorio_raw_min.json"
        ]
    },
    packages=['factorio_blueprint_analyser'],
)




import json

from factorio_blueprint_analyser import utils, config

# -----------------------------------------------------------
# Provide for the other files Factorio data
# from the factorio_blueprint_analyser/assets/factorio_raw/factorio_raw_min.json file
# -----------------------------------------------------------

recipies_key = "recipe"
recipies = {}

items_key = "item"
items = {}

entities_categories_keys = [
    "splitter",
    "container",
    "logistic-container",
    "assembling-machine",
    "infinity-container",
    "inserter",
    "underground-belt",
    "furnace",
    "transport-belt",
]
entities = {}


def load_data():
    global recipies, entities, items
    factorio_raw_data_file_path = config.config.data_file_path

    # TODO:. check that the file exists
    with open(factorio_raw_data_file_path, "r") as f:
        data = json.load(f)

        # Load the recipies
        if recipies_key not in data:
            utils.warning(f"Recipe key {recipies_key} not found in Factorio data")

        recipies = data[recipies_key]

        # Load the items
        if items_key not in data:
            utils.warning(f"Item key {items_key} not found in Factorio data")

        items = data[items_key]

        # Load the entities
        for key in entities_categories_keys:
            if key not in data:
                utils.warning(
                    f"Entity {key} category not found if Factorio data")
            else:
                for entity in data[key]:
                    entities[entity] = data[key][entity]

        utils.success(f"Factorio data successfully loaded")


def entity_exist(entity):
    # TODO
    pass
